
#include "vtkapp.h"
#include <stdlib.h>
#include <stdio.h>
//#include <conio.h>
#define SHEPARD_RESULT_TEST 0
#define SHEPARD_PRINT_SCREEN 0
#define X 0
#define Y 1
#define Z 2
#define W 3

#define RED 0
#define GREEN 1
#define BLUE 2
#define YELLOW 3
#define ALPHA 3



//------------------------------------
ImageData Illinois::data;
void Illinois::Initialize(void)
{
	cerr << "Opening Illinois Map Image" << endl;
	FILE* fin = fopen(MAP_IMAGEFILE,"r");
	if(fin == NULL)
	{
		cerr << "ERROR LOADING: " << MAP_IMAGEFILE << endl;
		exit(0);
	}
	
	cerr << "Allocating space" << endl; Illinois::data.pixels = new unsigned char[ MAP_HORIZ_RESOLUTION * MAP_VERT_RESOLUTION * 3];
	Illinois::data.width = MAP_HORIZ_RESOLUTION;
	Illinois::data.height = MAP_VERT_RESOLUTION;
	cerr << "Grabbing pixels" << endl;
	fread( Illinois::data.pixels, sizeof(unsigned char), MAP_HORIZ_RESOLUTION * MAP_VERT_RESOLUTION * 3, fin);
	cerr << "Closing file" << endl;
	fclose(fin);

}
///-----------------------------------
// VTK App messages supported
const int VTKApp::VTKAPP_KILL = -1;
const int VTKApp::VTKAPP_RENDER = 1;
//const int VTKApp::VTKAPP_BLEND_IMAGE = 2;
const int VTKApp::VTKAPP_1WEEK_SET = 3;
const int VTKApp::VTKAPP_SINGLEDAY_SET = 4;
const int VTKApp::VTKAPP_CREATE_IMAGEACTOR_SET = 5;
const int VTKApp::VTKAPP_CREATE_WINDDIR_GLYPHS_SET = 6;


//---------------------------------------
void ExitCallback(vtkFlRenderWindowInteractor* iren,void* data)
{
	cerr << "\"killing\" app" << endl;
	VTKApp* app = (VTKApp*) data;
	delete app;
}

//---------------------------------------
VTKApp::VTKApp(int x, int y, int width, int height, char* name, Fl_Window* pWindow):head(NULL),tail(NULL),killApp(false),lastPageFlip(0),currentActor(0)
{
	arrowCountPerDay = 0;
	numArrowDays = 0;
	renderWindow = vtkRenderWindow::New();
	renderWindow->SetSize(width,height);
	
	render1 = vtkOpenGLRenderer::New();
/*	
	if(pWindow)
	{
		cerr << "Adding to a valid fltk window" << endl;
		interactor = new vtkFlRenderWindowInteractor(x,y,width,height,NULL);
		pWindow->add( interactor );
		pWindow->end();
		pWindow->resizable( interactor );
		pWindow->show();
	}
	else
	{
		interactor = new vtkFlRenderWindowInteractor(x,y,width,height,name);
		interactor->label(name);
	}
	
	interactor->show();
	interactor->callback((Fl_Callback*)ExitCallback,(void*)this);
*/
	interactor = vtkRenderWindowInteractor::New();

//	renderWindow->BordersOff();
	render1->SetViewport( 0, 0, .75, 1);
	renderWindow->AddRenderer(render1);
	interactor->SetRenderWindow(renderWindow);
	interactor->Initialize();
	
	appName = new char[strlen(name) + 1];
	strcpy( appName, name);
	renderWindow->SetWindowName(appName);
	
	
	//----------------------More vtk pipeline stuff------------//
	oneWeekMode = false;
	timeSpanSet = false;
	actorsMade = false;
	canvases = NULL;
	canvasCount = 0;
	imageActors = NULL;
	imageCasts = NULL;
	imageMaps = NULL;
	imageLookup = NULL;
	imageLookupLookup = NULL;
	lookupCanvas = NULL;
	lookupActor = NULL;
	lookupCast = NULL;
	lookupMap = NULL;
	minTextActor = NULL;
	maxTextActor = NULL;
	currentDayActors = NULL;
	lookupRender = vtkRenderer::New();
	lookupRender->SetViewport(.75,0,1,1);
	lookupTranslate = vtkImageTranslateExtent::New();
	renderWindow->AddRenderer(lookupRender);
	scale[X] = scale[Y] = scale[Z] = 1;
	//---------------------------------------------------------//
	
	renderWindow->Render();

}

//---------------------------------------
/*
VTKApp::VTKApp(int x, int y, int width, int height, char* name, Fl_Double_Window*& pWindow):head(NULL),tail(NULL),killApp(false)
{
	renderWindow = vtkRenderWindow::New();
	renderWindow->SetSize(width,height);
	
	render1 = vtkOpenGLRenderer::New();
	interactor = new vtkFlRenderWindowInteractor(x,y,width,height,name);
	pWindow->add( interactor );
	pWindow->resizable( interactor );
	pWindow->show();
	interactor->show();
	

	renderWindow->AddRenderer(render1);
	interactor->SetRenderWindow(renderWindow);
	interactor->Initialize();
	
	appName = new char[strlen(name) + 1];
	strcpy( appName, name);


}
*/


//---------------------------------------
VTKApp::~VTKApp()
{		
	cerr << "VTKApp Destructor called" << endl;
	if(interactor)
		interactor->Delete();
	if(renderWindow)
		renderWindow->Delete();
	if(render1)
		render1->Delete();
	if(lookupRender)
		lookupRender->Delete();
		
	//destroy all messages
	VTKAppMessageQueue* temp;
	cerr << "Destroying Messages!" << endl;
	while( head )
	{
		temp = head;
		head = head->next;

		delete temp;
	}
			
			
	//------//
	cerr << "Number Of Days to Clean: " << canvasCount << endl;
	for(int i = 0; i < canvasCount; i++)
	{
		
		if(canvases[i])
			canvases[i]->Delete();
	
		if(imageActors[i])
			imageActors[i]->Delete();
			
		if(imageCasts[i])
			imageCasts[i]->Delete();
		
		if(imageMaps[i])
			imageMaps[i]->Delete();
			
		if(currentDayActors[i])
			currentDayActors[i]->Delete();
	}

	cerr << "Cleaning Out Arrays" << endl;
	if(imageLookupLookup)
		imageLookupLookup->Delete();
	if(canvases)
		delete [] canvases;
	if(imageActors)
		delete [] imageActors;
	if(imageCasts)
		delete [] imageCasts;
	if(imageMaps)
		delete [] imageMaps;
	if(imageLookup)
		imageLookup->Delete();
	if(lookupTranslate)
		lookupTranslate->Delete();
	if(currentDayActors)
		delete [] currentDayActors;
	if(minTextActor)
		minTextActor->Delete();
	if(maxTextActor)
		maxTextActor->Delete();
}


//---------------------------------------
char* VTKApp::GetAppName()
{
	return appName;
}

//---------------------------------------
void VTKApp::Render()
{

	if(renderWindow)
		renderWindow->Render();

}

//---------------------------------------
void VTKApp::Message(VTKAppMessage* vtkAppMsg)
{
	cerr << "App: " << GetAppName() << " Recieved Message : " << vtkAppMsg << endl;
	if(killApp || this == NULL)		//if we are killing or have killed the app, then don't recieve anymore messages
	{
		if( vtkAppMsg )
		{
//			if( vtkAppMsg->msgData )
	//			delete [] vtkAppMsg->msgData;
			
			delete vtkAppMsg;
		}
		
		return;	
	}
	
	VTKAppMessageQueue* newNode = new VTKAppMessageQueue();
	newNode->next = NULL;
	newNode->pMsgStruct = vtkAppMsg;
	//MessageBox(parentWindow,"Message Recieved","Acknowledgement",MB_OK);
	if(!head)
	{
		head = newNode;
		tail = newNode;
	}
	else
	{
		tail->next = newNode;
		tail = newNode;
		tail->next = NULL;
	}
}

//---------------------------------------
void VTKApp::ProcessMessages()
{
		
	while(head)
	{
		VTKAppMessageQueue* temp;
		float *sampDist = NULL;
		float *posTemp = NULL;
		float *tempFloatArray = NULL;
		double *tempDoubleArray = NULL;


		switch(head->pMsgStruct->msg)
		{
		case VTKApp::VTKAPP_KILL:
			//set the state to kill state
			killApp = true;
			//destroy all messages
			while( head )
			{
				temp = head;
				head = head->next;

//				if(temp->pMsgStruct->msgData != NULL)
	//				delete [] temp->pMsgStruct->msgData;

				delete temp;
			}
			
			//now the user must delete the object
			
		case VTKApp::VTKAPP_RENDER:
			//////////////////////////////////
			this->Render();
			break;
		case VTKApp::VTKAPP_1WEEK_SET:
			if( timeSpanSet == false )
			{
				oneWeekMode = true;
				timeSpanSet = true;
			}
			cerr << "App " << GetAppName() << " Set to 1 Week Mode" << endl;
			break;
		case VTKApp::VTKAPP_SINGLEDAY_SET:
			if( timeSpanSet == false )
			{
				oneWeekMode = false;
				timeSpanSet = true;
			}
			cerr << "App " << GetAppName() << " Set to Single Day Mode" << endl;
			break;
		case VTKApp::VTKAPP_CREATE_IMAGEACTOR_SET:
			cerr << "App " << GetAppName() << " Processing Image Generation" << endl;
			if(actorsMade == false)
			{
				CreateColorMapActors(head->pMsgStruct);
				actorsMade = true;
			}
			cerr << "App " << GetAppName() << " Generated Images" << endl;
			break;
		case VTKApp::VTKAPP_CREATE_WINDDIR_GLYPHS_SET:
			if(actorsMade == false)
			{
				CreateWindDirectionGlyphs(head->pMsgStruct);
				actorsMade = true;
			}
			break;
//		case VTKApp::VTKAPP_BLEND_IMAGE:
//			BlendImage(head->pMsgStruct);
//			break;
		default:
			//////////////////////////////////
			//nothing to process
			break;
		};
		
		temp = head;
		head = head->next;

//		if(temp->pMsgStruct->msgData != NULL)
	//		delete [] temp->pMsgStruct->msgData;

		delete temp;
	}

}

//---------------------------------------
bool VTKApp::HaveMessages()
{
	if( head != NULL)
		return true;
	else
		return false;
}

//---------------------------------------
bool VTKApp::KillApp()
{
	return killApp;
}

//---------------------------------------
void VTKApp::BlendImage(void* data)
{
	

}


//----------------------------------
bool VTKApp::Animation()
{
	return oneWeekMode;
}


//----------------------------------
void VTKApp::CycleActors()
{		
		if(clock() - lastPageFlip >= 1000)
		{
			//hide the current actor...show the next actor
		//	render1->RemoveActor( actors[currentActor]);
		//	currentActor = (currentActor + 1) % canvasCount;
		//	render1->AddActor( actors[currentActor] );
		//	render1->ResetCamera();
			render1->RemoveActor( imageActors[currentActor]);
			render1->RemoveActor2D( currentDayActors[currentActor]);
			
			currentActor = (currentActor + 1) % canvasCount;
			
			render1->AddActor( imageActors[currentActor] );
			render1->AddActor2D( currentDayActors[currentActor]);
			render1->ResetCamera();
			Render();
			lastPageFlip = clock();
		}	
	
}

