#ifndef VTK_APP_H
#define VTK_APP_H
//#include <QUANTA/QUANTA.hxx>
#include <pthread.h>

#ifdef WIN32
#include <vtk/vtkRenderWindow.h>
#include <vtk/vtkRenderer.h>
#include <vtk/vtkOpenGLRenderer.h>
#include <vtk/vtkRenderWindowInteractor.h>
#include <vtk/vtkConeSource.h>
#include <vtk/vtkPolyDataMapper.h>
#include <vtk/vtkActor.h>
#include <vtk/vtkProperty.h>
#include <vtk/vtkCamera.h>
#include <vtk/vtkInteractorStyleTrackball.h>

//--------------------------------------
#include <vtk/vtkImageReader2.h>
#include <vtk/vtkLODActor.h>
#include <vtk/vtkLookupTable.h>
#include <vtk/vtkContourFilter.h>
#include <vtk/vtkImageClip.h>
#include <vtk/vtkGeometryFilter.h>
#include <vtk/vtkMarchingContourFilter.h>
#include <vtk/vtkExtractVOI.h>

//---------------------------------------
#include <vtk/vtkPiecewiseFunction.h>
#include <vtk/vtkColorTransferFunction.h>
#include <vtk/vtkVolumeProperty.h>
#include <vtk/vtkVolumeRayCastCompositeFunction.h>
#include <vtk/vtkVolumeRayCastMIPFunction.h>
#include <vtk/vtkVolumeRayCastMapper.h>
#include <vtk/vtkVolume.h>
#include <vtk/vtkImageCanvasSource2D.h>
#include <vtk/vtkPlaneSource.h>
#include <vtk/vtkTexture.h>
#else

#include <vtkRenderWindow.h>
#include <vtkRenderer.h>
#include <vtkOpenGLRenderer.h>
#include <vtkRenderWindowInteractor.h>
#include <vtkConeSource.h>
#include <vtkPolyDataMapper.h>
#include <vtkActor.h>
#include <vtkProperty.h>
#include <vtkCamera.h>
#include <vtkInteractorStyleTrackball.h>

//--------------------------------------
#include <vtkImageReader2.h>
#include <vtkTexture.h>
#include <vtkLODActor.h>
#include <vtkLookupTable.h>
#include <vtkContourFilter.h>
#include <vtkImageClip.h>
#include <vtkGeometryFilter.h>
#include <vtkMarchingContourFilter.h>
#include <vtkExtractVOI.h>

//---------------------------------------
#include <vtkPiecewiseFunction.h>
#include <vtkColorTransferFunction.h>
#include <vtkVolumeProperty.h>
#include <vtkVolumeRayCastCompositeFunction.h>
#include <vtkVolumeRayCastMIPFunction.h>
#include <vtkVolumeRayCastMapper.h>
#include <vtkVolume.h>

//-------------------------------------
#include <vtkImageCanvasSource2D.h>
#include <vtkPlaneSource.h>
#include <vtkTextActor.h>
#include <vtkImageActor.h>
#include <vtkImageCast.h>
#include <vtkArrowSource.h>
#include <vtkImageMapToColors.h>
#include <vtkImageMapper.h>
#include <vtkImageTranslateExtent.h>

#endif

#include "vtkFlRenderWindowInteractor.h"
#include "bigDataLoader.hxx"
#include <FL/Fl_Window.H>
#include <FL/Fl_Double_Window.H>


//--------------------------------------
//The degrees latitude and longitude based on the change in the respective directions
//And other Map related info!
#define DEG_LAT_PER_DY 0.0096f
#define DEG_LONG_PER_DX 0.0124f
#define DEG_LAT_AT_ZERO 87.636f
#define DEG_LONG_AT_ZERO 36.6944f
#define MAP_HORIZ_RESOLUTION 338
#define MAP_VERT_RESOLUTION 599
#define MAP_IMAGEFILE "./data/illinois.raw"
#define LAST_YEAR 2003
//#define WIND_HORIZ_RESOLUTION 20
//#define WIND_VERT_RESOLUTION 35
#define WIND_RESOLUTION 0.01f
#define ARROW_TIP_RESOLUTION 1
#define ARROW_SHAFT_RESOLUTION 2
//--------------------------------
//Global Data!!!!!!
extern WeatherService* gWeatherService;
//--------------------------------------
//Defined Attributes for InterestedDay structures' attribute field

struct InterestedDay
{
	int month;
	int day;
	int year;
	int attribute;
	WeatherService* pWS;
};

struct ContourData
{
	int color;
	int level;
	int transparency;
};

struct LookupTableData
{
	float *thresholds;	//The values inbetween the thresholds, [threshold[n], threshold[n+1]),
				//will fall into that particular "bin"
				//The values are "real world" data values (e.g. Humidity)
	int thresholdCount;
	unsigned char** tableValues;	//2D array of values each row maps to 
					//the ranges between thresholds
					//and the row contains rgb values to color map	
	
};

struct GreyLookupTableData
{
	float *thresholds;	//The values inbetween the thresholds, [threshold[n], threshold[n+1]),
				//will fall into that particular "bin"
				//The values are "real world" data values (e.g. Humidity)
	int thresholdCount;
	unsigned char* tableValues;	//grey scale values
	
};

struct ImageData
{
	unsigned char* pixels;
	int width;
	int height;
	LookupTableData lupTable;

};

struct VTKAppMessage
{
	int	msg;
	void* msgData;	//assume the data is always in array form!  Single values are from arrays of size 1!
};

struct VTKAppMessageQueue
{
	VTKAppMessage* pMsgStruct;
	VTKAppMessageQueue* next;
};

//-------------------------------------
class VTKApp;
//---------------------------------------



//		VTK APPLICATION CLASS
//		- Change the Render method to what ever way you want
//----------------------------------
void ExitCallback(vtkFlRenderWindowInteractor* iren,void* data);

//class used to hold the Map Image of Illinois
class Illinois
{
private:
	friend class VTKApp;
	static ImageData data;
public:
	static void Initialize(void);
};

//Prototype for routine that is used to cycle/"animate" the actors
void* ThreadApp(void*);
void* cycleThread(void* );

class VTKApp 
{
public:
	VTKApp(int, int, int, int, char*,Fl_Window* );
	virtual ~VTKApp();
	
	virtual void	Render();
	void	Message(VTKAppMessage* );	//recieve messages
	void	ProcessMessages();			//process messages in message queue
	bool	HaveMessages();
	char*	GetAppName();
	bool 	KillApp();
	bool	Animation();
	void	CycleActors();
	
//	void	GetThread(QUANTAts_thread_c*& );
//	void	SetThread(QUANTAts_thread_c* );
	///-----------------------------------
	// VTK App messages supported
	static const int VTKAPP_KILL;
	static const int VTKAPP_RENDER;
	// static const int VTKAPP_BLEND_IMAGE;	//Blend the following image onto all the images!!!!
	

	// App messages to signify if single day actorset or 2 week animated set

	//---------------------------------WARNING!!!!------------------------------------//
	//	These must be sent before sending the visual representations that follow!!//
	//--------------------------------------------------------------------------------//
	static const int VTKAPP_1WEEK_SET;
	static const int VTKAPP_SINGLEDAY_SET;
	
	// App messages to signify if it's an image or 3d/glyph representation
	static const int VTKAPP_CREATE_IMAGEACTOR_SET;
	static const int VTKAPP_CREATE_WINDDIR_GLYPHS_SET;
	

private:
	friend void ExitCallback(vtkFlRenderWindowInteractor* iren,void* data);
//	friend void* cycleThread(void*);
//	friend void* ThreadApp(void*);
//	bool	Updated();
	char* appName;
	bool  killApp;
	VTKAppMessageQueue *head;
	VTKAppMessageQueue *tail;

	vtkOpenGLRenderer		*render1;
	vtkRenderWindow			*renderWindow;
	//vtkFlRenderWindowInteractor	*interactor;
	vtkRenderWindowInteractor	*interactor;
	vtkImageCanvasSource2D		**canvases;
	int				canvasCount;	//the number of canvases and
					//actors = canvasCount

	vtkImageActor			**imageActors;
	vtkImageCast			**imageCasts;
	vtkImageMapToColors		**imageMaps;
	vtkLookupTable			*imageLookup;
	vtkLookupTable			*imageLookupLookup;
	vtkImageTranslateExtent		*lookupTranslate;
	vtkRenderer			*lookupRender;
	//--------Wind Glyphs
	int				arrowCountPerDay;
	int				numArrowDays;
	int				dailyArrowRowCnt;
	int				dailyArrowColCnt;
		
	vtkImageCanvasSource2D		*lookupCanvas;
	vtkImageActor			*lookupActor;
	vtkImageCast			*lookupCast;
	vtkImageMapToColors		*lookupMap;
	
	vtkTextActor			*minTextActor;
	vtkTextActor			*maxTextActor;
	vtkTextActor			**currentDayActors;
	
	//scales to stretch the planes
	float				scale[3];
	
	bool				oneWeekMode;
	bool				timeSpanSet;
	bool				actorsMade;
	int				currentActor;
	clock_t				lastPageFlip;
	//Private Methods to get the job done!
	void BlendImage(void*);
	void CreateWindDirectionGlyphs(void*);
	void CreateColorMapActors(void*);
//	QUANTAts_thread_c	thisThread;
};


struct VTKAppProperties
{
	int startx;
	int starty;
	int width;
	int height;
	char* name;
	VTKApp** ppApp;
	InterestedDay day;
};





#endif
