
#include "vtkapp.h"
#include <vtkCamera.h>
#include <stdlib.h>
#include <stdio.h>
//#include <conio.h>
#define SHEPARD_RESULT_TEST 0
#define SHEPARD_PRINT_SCREEN 0
#define X 0
#define Y 1
#define Z 2
#define W 3

#define RED 0
#define GREEN 1
#define BLUE 2
#define YELLOW 3
#define ALPHA 3
//--------------------------------------=
char* NumToChar(int num){
static int length;
	
	length = 0;
	
	if(num == 0)
		return "0";

	for(int temp = num; temp; temp /= 10)
		length++;
		
	char *text = new char[length + 1];
	
	text[length] = '\0';
	
	//printf("text: %s\n", text);
	
	for(int i = length - 1; i >= 0; i--, num /= 10)
		text[i] = num % 10 + 48;
	
	//printf("text: %s\n", text);
		
	return text;
}
//---------------------------------------
void VTKApp::CreateColorMapActors(void* data)
{
	VTKAppMessage* msg = (VTKAppMessage* )data;
	InterestedDay* interest = (InterestedDay*) msg->msgData;
	cerr << "looking for date: " << interest->month << ", "  << interest->day << ", " << interest->year << endl;
	
	//do an error check
	switch(interest->attribute)
	{
	case AVG_WIND_SPEED:
	case AVG_WIND_DIRECTION:
	case AVG_AIR_TEMP:
	case AVG_SOIL_TEMP:
	case AVG_HUMIDTY:
	case TOTAL_PRECIPITATION:
		break;
	default:
		//shouldn't ever happen!
		return;
		break;
	}
	
	//perform shepards method and generate color map to map the colors when we draw it out!
	//also depends on the number of days!...if 14 days exceeds the data available we will cut it short!
	int numberOfDays = 1;
	if(oneWeekMode)
	{
		numberOfDays = 7;
		if( interest->year == LAST_YEAR && interest->month == 12 && (interest->day + 7) > 31)
		{
			numberOfDays -= (interest->day + 7) - 31; 
		}
	}
	
	//now we have the number of days!
	//allocate objects aswell!
	canvases = new vtkImageCanvasSource2D*[numberOfDays];
	canvasCount = numberOfDays;
//	imageCasts = new vtkImageCast*[numberOfDays];
	int dayItr = 0;
	int month = interest->month;
	int day = interest->day;
	int year = interest->year;
	
	int itr;	
	
	float **values;
	values = new float*[numberOfDays];
	for( itr = 0; itr < numberOfDays; itr++)
	{
		values[itr] = new float[ MAP_HORIZ_RESOLUTION * MAP_VERT_RESOLUTION ];
	
	}
	
	//Range Variables Used Later For Lookup Table Values	
	float MinValue,MaxValue;
	MinValue = 1000;
	MaxValue = -1000;
	int pixelX, pixelY;
	//---------------------------
#if SHEPARD_RESULT_TEST
//		FILE* fout = fopen("results.raw","wb");
		FILE* fout2 = fopen("tableMaps.txt","w");
#endif
	
	float illegalMax = 999.9;
	
	while( dayItr < numberOfDays)
	{
		int length = 0;
		
		if(dayItr > 0)
			GetNextDate(month,day,year);
		
		DayInFile* daysArray = NULL;
		cerr << "Global Weather Service Object at @ " << gWeatherService << " -- WeatherService Sent to Method @: " << interest->pWS << endl;
		WeatherService* ws = gWeatherService;
		
		if( interest->pWS != NULL)
		{
			ws = interest->pWS;
			daysArray = GetDayInFileArray( month, day, year,  interest->pWS, length);
		}
		else	
			daysArray = GetDayInFileArray( month, day, year, gWeatherService, length);
		float** attributeValues = new float*[length];
		cerr << "DayArray Address @ " << daysArray << " Number of Scatter Points: " << length << endl;
		//Allocate the actual vtk image canvas object
		canvases[dayItr] = vtkImageCanvasSource2D::New();
		
		//get the attributes we are interested in!
		int i;
		cerr << "Attribute values: " << endl;;
		for(i = 0; i < length; i++)
		{
			switch(interest->attribute)
			{
			case AVG_WIND_SPEED:
				attributeValues[i] = &(daysArray[i].avgWindSpeed);
				break;
			case AVG_WIND_DIRECTION:
				attributeValues[i] = &(daysArray[i].avgWindDir);
				break;
			case AVG_AIR_TEMP:
				attributeValues[i] = &(daysArray[i].avgAirTemp);
				break;
			case AVG_SOIL_TEMP:
				attributeValues[i] = &(daysArray[i].avgSoilTemp);
				break;
			case AVG_HUMIDTY:
				attributeValues[i] = &(daysArray[i].avgHumidity);
				break;
			case TOTAL_PRECIPITATION:
				attributeValues[i] = &(daysArray[i].totalPrecip);
				illegalMax = 99.9;
				break;
			}//------------------------End switch based on interest->attribute
			
			cerr << *(attributeValues[i]) << " ";
			if( (i + 1) % 5 == 0)
				cerr << endl;
		} //----------------------End for loop to loop through the daysArray
		cerr << endl;
		
		//-----------------------Begin interpolating and drawing the texture
		canvases[dayItr]->SetExtent(0,MAP_HORIZ_RESOLUTION - 1,0,MAP_VERT_RESOLUTION -1, 0, 0);
		cerr << "Starting Inverse Distance Interpolation @ " << clock() << endl;
		//now we actuall run the shepard method!

		
		for(pixelY = 0; pixelY < MAP_VERT_RESOLUTION; pixelY++)
		{
			for(pixelX = 0; pixelX < MAP_HORIZ_RESOLUTION; pixelX++)
			{
#if SHEPARD_PRINT_SCREEN
				cerr << "pixel " << pixelX << ", " << pixelY;
#endif
				double* distance = NULL;
				double maxDistance;
				double pixelLongitude, pixelLatitude;
				double* weights = NULL;
				if(!weights) 
					weights = new double[length];
					
				pixelLongitude = (double)pixelX * (double)DEG_LONG_PER_DX + (double)DEG_LONG_AT_ZERO;
				pixelLatitude = (double)pixelY * (double)DEG_LAT_PER_DY + (double)DEG_LAT_AT_ZERO;
				maxDistance = 0;
				int k;
#if SHEPARD_PRINT_SCREEN
				cerr << " pixel " << pixelLongitude << "," << pixelLatitude << " Getting Max Distance - ";
#endif
				//get distance to farthest scatter point
				if(!distance) 
					distance = new double[length];
					
				for( k = 0; k < length; k++)
				{	
#if SHEPARD_PRINT_SCREEN
					cerr << endl <<  "scatter point " << ws->stationListing[k]->longitude << ',' << ws->stationListing[k]->latitude << ' ';
#endif
					distance[k] = Distance(pixelLongitude,pixelLatitude,ws->stationListing[k]->longitude,
								ws->stationListing[k]->latitude);
#if SHEPARD_PRINT_SCREEN
					cerr << '?';
#endif
					if(distance[k] > maxDistance )
						maxDistance = distance[k];
					
				}
#if SHEPARD_PRINT_SCREEN
				cerr << "Applying weights to value - ";
#endif
				//apply the weights to each attribute value!
				double value;
				double numerator, denominator;
				value = 0;
				numerator = 0;
				denominator = 0;
				values[dayItr][pixelX + (pixelY * MAP_HORIZ_RESOLUTION)] = 0;
				for( k = 0; k < length; k++)
				{
#if SHEPARD_PRINT_SCREEN
					cerr << "attribute value: " << *(attributeValues[k]) << " - weight: " << weights[k] << " result: " << (weights[k] * (*(attributeValues[k]))) << endl;
#endif
					if((*attributeValues[k]) < illegalMax )	// && (*attributeValues[k]) != 999.9)
					{
						//value += (weights[k]) * (*(attributeValues[k]));
						numerator += ((*attributeValues[k]) / (distance[k] * distance[k] * distance[k]));		
						denominator += (1.0f / (distance[k] * distance[k] * distance[k]));
					}
				}
				
				//value = (numerator / denominator);
				values[dayItr][pixelX + (pixelY * MAP_HORIZ_RESOLUTION)] = numerator / denominator;
				
			
			//	value /= (double)length;
				//Now we have our interpolated value!!
				//We need to map it to a color!!
#if SHEPARD_PRINT_SCREEN
				cerr << value << " Min Max Check -" << endl;
#endif
				if( values[dayItr][pixelX + (pixelY * MAP_HORIZ_RESOLUTION)] < illegalMax )// &&
				//	values[dayItr][pixelX + (pixelY * MAP_HORIZ_RESOLUTION)] != 999.9 )
				{
					if(values[dayItr][pixelX + (pixelY * MAP_HORIZ_RESOLUTION)] <= MinValue )
					{
					//	cerr << "New Minimum: " << values[dayItr][pixelX + (pixelY * MAP_HORIZ_RESOLUTION)];
						MinValue = values[dayItr][pixelX + (pixelY * MAP_HORIZ_RESOLUTION)];
					}
					if(values[dayItr][pixelX + (pixelY * MAP_HORIZ_RESOLUTION)] >= MaxValue)
					{
					//	cerr << "New Maximum: " << values[dayItr][pixelX + (pixelY * MAP_HORIZ_RESOLUTION)];
						MaxValue = values[dayItr][pixelX + (pixelY * MAP_HORIZ_RESOLUTION)];
					}
				}	
				i = 0;
#if SHEPARD_RESULT_TEST
				//fprintf(fout,"%d %d %d ",lookup.tableValues[k][0],lookup.tableValues[k][1],lookup.tableValues[k][2]);
				//cerr << "pixel " << pixelX << ", " << pixelY << endl;
//				fwrite(lookup.tableValues[k],sizeof(unsigned char),3,fout);


				fprintf(fout2,"%d,%d : value: %f @ %d ", pixelX, pixelY, 
					values[dayItr][pixelX + (pixelY * MAP_HORIZ_RESOLUTION)], &(values[dayItr][pixelX + (pixelY * MAP_HORIZ_RESOLUTION)]) );
					
				fprintf(fout2,"\n");
#endif
			
			}//--------------------End PixelX loop
#if SHEPARD_RESULT_TEST
		fprintf(fout2,"\n");
#endif
			cerr << ".";
			
			
		}//------------------End PixelY loop
		cerr << " --- Finished @ " << clock() << endl;
	
		dayItr++;
	}//---------------------------End While Loop for each day to process the values of that day

#if SHEPARD_RESULT_TEST
		cerr << "Closing results file...";
//		fflush(fout);
		fflush(fout2);
//		fclose(fout);
		fclose(fout2);
		cerr << "FILE CLOSED" <<  endl;
#endif

	//------------------------Generate the lookup table for creating textures based on attribute properties and Range of Values!
	GreyLookupTableData lookup;
	float base_offset = 0;
	float interval;
	interval = (float)(MaxValue - MinValue) / 9.0f;	//We always want 6 different ranges
	lookup.thresholdCount = 10;
	lookup.thresholds = new float[lookup.thresholdCount];
	base_offset = MinValue;
	imageLookup = vtkLookupTable::New();
	imageLookup->SetTableRange( 0, 256 );
	imageLookup->SetHueRange( 0, 1);
	imageLookup->SetSaturationRange(0, 1);
	imageLookup->SetValueRange( 1, 1);
	imageLookup->Build();
	imageLookup->SetTableValue(255,1,0,0);
	imageLookup->SetTableValue(254,.2,.2,.2);
	
	float rgb[3];
	FILE* colorOut = fopen("colors.txt","w");
	for( itr = 0; itr < 256; itr++)
	{
		imageLookup->GetColor(itr,rgb);
		fprintf(colorOut,"map %d: -> rgb: %f , %f , %f\n", itr, rgb[0],rgb[1],rgb[2]);
		
	}
	
	fclose(colorOut);
	

	lookup.tableValues = new unsigned char[lookup.thresholdCount];
	
	for( itr = 0; itr < lookup.thresholdCount; itr++)
	{
		lookup.thresholds[itr] = itr * interval + base_offset;
		lookup.tableValues[itr] = ( 255 / lookup.thresholdCount ) * itr;	
	
	}
//#if SHEPARD_PRINT_SCREEN	
	cerr << "Min Value: " << MinValue << " --- MaxValue: " << MaxValue << " ---- Interval: " << interval << endl;
	for(itr = 0; itr < lookup.thresholdCount; itr++)
	{
		imageLookup->GetColor( (int)lookup.tableValues[itr], rgb );
		cerr << "lookup threshold # " << itr << " has value: " << lookup.thresholds[itr] << " with color map value of " 
			<< (int)lookup.tableValues[itr] << " and RGB of: " 
			<< rgb[0] << ", " << rgb[1] << ", " << rgb[2] << endl;  //[0] << ',' << (int)lookup.tableValues[itr][1] << ',' <<  (int)lookup.tableValues[itr][2] << endl;		
	
	}
//#endif

#if SHEPARD_RESULT_TEST
	FILE* fileOut;
	FILE* tableOut;
#endif	
	//**********************************************
	//Now loop through the canvases and draw them out!!!!
	for( dayItr = 0; dayItr < numberOfDays; dayItr++)
	{
#if SHEPARD_RESULT_TEST
		char buf[12];
		buf[0] = dayItr + 48;
		buf[1] = '\0';
		strcat(buf,".raw");
		fileOut = fopen(buf,"wb");
		buf[1] = '\0';
		strcat(buf,".txt");
		tableOut = fopen(buf,"w");
		
#endif		
		cerr << endl << "Generating Texture for Day " << dayItr << endl;
		for(pixelY = 0; pixelY < MAP_VERT_RESOLUTION; pixelY++)
		{
			for(pixelX = 0; pixelX < MAP_HORIZ_RESOLUTION; pixelX++)
			{
				int map;
				for( map = 0; map < lookup.thresholdCount; map++)
				{
					if(values[dayItr][pixelX + (pixelY * MAP_HORIZ_RESOLUTION)] < lookup.thresholds[map])
					{
						break;	
					}
					
					
				}
				
				if(map < 0)
					map = 0;
				else if(map >= lookup.thresholdCount)
					map = lookup.thresholdCount - 1;
#if SHEPARD_RESULT_TEST
				fprintf(tableOut,"mapped %d - ", map);	
				fflush(tableOut);
#endif		
				//BAKING OF ILLINOIS IMAGE if the color is a certain level
				unsigned char bakeColor[3];
				memcpy(bakeColor , &Illinois::data.pixels[ (pixelX * 3) + (pixelY * MAP_HORIZ_RESOLUTION * 3)], sizeof(unsigned char) * 3);
				bool bake;
				bake = false;
				if( bakeColor[1] < 70 || bakeColor[2] < 70)
				{ 
					//increase the darkness 
					bake = true;
#if SHEPARD_RESULT_TEST	
					fprintf(tableOut,"Baking! %d %d %d :::: ", bakeColor[0], bakeColor[1], bakeColor[2]);
#endif				
					memset(bakeColor,0,sizeof(unsigned char) * 3);	
					canvases[dayItr]->SetDrawColor( 255 );
				}
				else
				{
					//canvases[dayItr]->SetDrawColor( lookup.tableValues[map][0], lookup.tableValues[map][1], lookup.tableValues[map][2], 0);
					canvases[dayItr]->SetDrawColor(lookup.tableValues[map]);
				}
				
				canvases[dayItr]->DrawPoint( pixelX, MAP_VERT_RESOLUTION - pixelY );
#if SHEPARD_RESULT_TEST
				fprintf(tableOut,"drawn - ");
				fflush(tableOut);
				if(!bake)
					fwrite(lookup.tableValues[map],sizeof(unsigned char),1,fileOut);
				else
					fwrite(bakeColor,sizeof(unsigned char),3,fileOut);
				fprintf(tableOut,"%d, %d - %d - %f -> %d\n", pixelX, pixelY, 
					&( values[dayItr][pixelX + (pixelY * MAP_HORIZ_RESOLUTION)]), values[dayItr][pixelX + (pixelY * MAP_HORIZ_RESOLUTION)], map);
#endif				
			
			}	//End for each PixelX
			cerr << '.';
		}//End for each PixelY	
#if SHEPARD_RESULT_TEST
		fflush(fileOut);
		fclose(fileOut);
		fflush(tableOut);
		fclose(tableOut);
#endif
		cerr << "Cleaning out 2D Array of Interpolated Values for Day " << dayItr << endl;
		delete [] values[dayItr];
		values[dayItr] = NULL;
			
	}
	
	delete [] values;
	


	//_____________________________________VTK PIPELINE CREATION!!!!________________________//
	imageActors = new vtkImageActor*[numberOfDays];
	imageCasts = new vtkImageCast*[numberOfDays];
	imageMaps = new vtkImageMapToColors*[numberOfDays];
	
	
	for( dayItr = 0; dayItr < numberOfDays; dayItr++)
	{
		
		//----------GOOOD CODE!!!!
		imageActors[dayItr] = vtkImageActor::New();
		imageCasts[dayItr] = vtkImageCast::New();
		imageMaps[dayItr] = vtkImageMapToColors::New();
		
		imageCasts[dayItr]->SetInput( canvases[dayItr]->GetOutput() );
		imageCasts[dayItr]->SetOutputScalarTypeToUnsignedChar();
		
		imageMaps[dayItr]->SetInput( imageCasts[dayItr]->GetOutput() );
		imageMaps[dayItr]->SetLookupTable( imageLookup );
		imageMaps[dayItr]->SetOutputFormatToRGB();
		
		imageActors[dayItr]->SetInput( imageMaps[dayItr]->GetOutput() );
		//imageActors[dayItr]->SetDisplayExtent(0,337,0,598,0,0);
		//_________________
		
	}
	

	render1->AddActor( imageActors[0] );
	//now we add the lookuptable pipeline
	lookupActor = vtkImageActor::New();
	lookupCast = vtkImageCast::New();
	lookupMap = vtkImageMapToColors::New();
	lookupCanvas = vtkImageCanvasSource2D::New();
	imageLookupLookup = vtkLookupTable::New();
	imageLookupLookup->SetTableRange( 0, 256 );
	imageLookupLookup->SetHueRange( 0, 1);
	imageLookupLookup->SetSaturationRange(0, 1);
	imageLookupLookup->SetValueRange( 1, 1);
	imageLookupLookup->Build();
	
	lookupCanvas->SetExtent(0,10,0,MAP_VERT_RESOLUTION,0,0);
	for( int map = 0; map < lookup.thresholdCount; map++)
	{
		lookupCanvas->SetDrawColor( lookup.tableValues[map]);
		if( map + 1 < lookup.thresholdCount)
		{
			lookupCanvas->FillBox( 0, 10 + 10, map * (MAP_VERT_RESOLUTION / 10) , 
				map * (MAP_VERT_RESOLUTION / 10 ) + (MAP_VERT_RESOLUTION / 10 ));
		}
		else
		{
			lookupCanvas->FillBox( 0, 10 + 10, map * (MAP_VERT_RESOLUTION / 10) , MAP_VERT_RESOLUTION);
		
		}
			
	}
	
	
	lookupCast->SetInput( lookupCanvas->GetOutput() );
	lookupCast->SetOutputScalarTypeToUnsignedChar();
	
	lookupMap->SetInput( lookupCast->GetOutput() );
	lookupMap->SetLookupTable( imageLookupLookup );
	lookupMap->SetOutputFormatToRGB();
	lookupActor->SetInput( lookupMap->GetOutput() );
	lookupRender->AddActor( lookupActor );

	//---------Add Text
	//First the Date(s)
	currentDayActors = new vtkTextActor*[numberOfDays];
	char buffer[64] = "";
	month = interest->month;
	day = interest->day;
	year = interest->year;
	
	for( itr = 0; itr < numberOfDays; itr++)
	{
		currentDayActors[itr] = vtkTextActor::New();
		buffer[0] = '\0';
		strcat(buffer,"Date: ");
		strcat(buffer, NumToChar(month));
		strcat(buffer,"/");
		strcat(buffer, NumToChar(day));
		strcat(buffer,"/");
		strcat(buffer, NumToChar(year));
		
		GetNextDate(month,day,year);
		
		
		currentDayActors[itr]->SetInput(buffer);
		currentDayActors[itr]->SetPosition( 10, 10 );
		currentDayActors[itr]->SetHeight(15);
		
	}
	
	//Now the Min & Max Values
	minTextActor = vtkTextActor::New();
	maxTextActor = NULL; //vtkTextActor::New();
	buffer[0] = '\0';
	char units[8] = "";
	
	switch(interest->attribute)
	{
	case AVG_WIND_SPEED:
		strcat(units,"\nMPH");
		break;
	case AVG_WIND_DIRECTION:
		break;
	case AVG_AIR_TEMP:
	case AVG_SOIL_TEMP:
		strcat(units,"\nDEG F");
		break;
	case AVG_HUMIDTY:
		strcat(units,"\n%");
		break;
	case TOTAL_PRECIPITATION:
		strcat(units,"\nINCHES");
		break;
	default:
		//shouldn't ever happen!
		return;
		break;
	}
	
	//setup strings to actors
	char maxString[8] = "";
	char minString[8] = "";
	sprintf(minString, "%'.2f",MinValue);
	sprintf(maxString,"%'.2f",MaxValue);
	strcat(buffer,"Max:");
	strcat(buffer, maxString);
	strcat(buffer, "\n\n\n");
	strcat(buffer,"Min:");
	strcat(buffer, minString);
	strcat(buffer,"\n\nUnits: ");
	strcat(buffer, units);
	minTextActor->SetInput(buffer);
	//buffer[0] = '\0';
	//strcat(buffer, NumToChar((int)MaxValue));
	//strcat(buffer, units);
	//maxTextActor->SetInput(buffer);
	
	//position the actors
	minTextActor->SetPosition( 0, 10 );
	
	render1->AddActor2D( currentDayActors[0]);
	render1->ResetCamera();
	lookupRender->AddActor2D(minTextActor);
	lookupRender->ResetCamera();
	Render();
	canvasCount = numberOfDays;
	currentActor = 0;
	//pthread_create(&cycleThreadID, NULL, cycleThread, (void*)(this)); 
	//cycleThread((void*)this);
	lastPageFlip = clock();
	
}	//END OF COLORMAP ACTOR CREATION METHOD
