#include <pthread.h>
#include "vtkappList.h"
#include "vtkFlRenderWindowInteractor.h"

VTKAppList::VTKAppList()
{
	release = false;
	freed = false;
	NoAppRequested = true;
	pthread_mutex_destroy( &messageMutex );
	PauseAnims = false;
}

VTKAppList::~VTKAppList()
{
	KillAllApps();
}

//---------------------
//

void* AppListThread(void* data)
{
	cerr << "AppList Thread Beginning!" << endl;
	VTKAppList** aList = (VTKAppList**)data;
	(*aList) = new VTKAppList();
	
	(*aList)->Initialize();
	while((*aList)->NoAppRequested);
	
	while(!(*aList)->Release())
	{
		//cerr << "Top of AppList Thread Loop" << endl;
		//start any new apps
		if(  (*aList)->executeAppQueue || (*aList)->executeQueue.size() )
		{
			unsigned int ele;
			ele = 0;
			cerr << "Apps To Execute Pending" << endl;
		//	(*aList)->WaitForExecuteAppAccess();
			while( (*aList)->executeQueue.size())
			{
				list<VTKAppProperties* >::iterator itr = (*aList)->executeQueue.begin();
				cerr << "Found a New App To Start" << endl;
				VTKAppProperties* temp = *(itr);
				cerr << "App properties at: " << *itr << endl;
				cerr << "App Properties are: " << temp->startx << ',' << temp->starty
					<< " @ " << temp->width << 'x' << temp->height << ' ' << temp->name << endl;
				(*(temp->ppApp))= new VTKApp( temp->startx, temp->starty,
								temp->width, temp->height,
								temp->name,  NULL);
				
				cerr << "App Pointer at: " << temp->ppApp << " pointing to App at: " << *(temp->ppApp) << endl;
				
				//temp->ppApp = NULL;
				//delete temp;
				(*aList)->executeQueue.pop_front();
				ele++;
				//(*(temp->ppApp))->Render();
				//itr++;
			}
			
			(*aList)->executeQueue.clear();
			cerr << "Finished Making New Renderings" << endl;
		//	(*aList)->FreeExecuteAppAccess();
		}//end if there are new apps to start
		(*aList)->executeAppQueue = false;
		//cerr << '&';
		int i;
		i = (*aList)->GetNumApps();
		list<VTKAppRelayStation* >::iterator itr = (*aList)->listing.begin();
		clock_t loopstart;
		loopstart = clock();
		while( i > 0)
		{
			//cerr << "Looping through apps started at: " << loopstart << endl;
			//----------------------Send any relay Messages to the App
			VTKAppRelayStation* appRS = (*itr);
			VTKApp* app = *(appRS->ppApp);  //(*aList)->GetApp(i);
			if(!app)
				continue;
				
			VTKAppMessageQueue* tempMsgQueue;
			VTKAppMessage* tempMsg;
			if( appRS->msgQueueHead)
			{
				cerr << "messages to be relayed to app!" << endl;
			
			//	(*aList)->WaitForMessageQueueAccess();	
					while( appRS->msgQueueHead)
					{	
						tempMsgQueue = appRS->msgQueueHead;
						appRS->msgQueueHead = appRS->msgQueueHead->next;
						app->Message( tempMsgQueue->pMsgStruct );
						tempMsgQueue->pMsgStruct = NULL;
						delete tempMsgQueue;
					}
			//	(*aList)->FreeMessageQueueAccess();
			}
			
			//---------------------------
			if(app)
			{
				if(app->HaveMessages())
					app->ProcessMessages();

				if(!(*aList)->PauseAnims && app->Animation())
					app->CycleActors();	//cycle actors takes care of the page flipping time intervals

				app->Render();

			}//done processing messages for the app
			
			itr++;
			i--;
			
		}//done looping through the apps
	}//end thread

	(*aList)->KillAllApps();
	(*aList)->freed = true;
	return (void*)NULL;

}


//---------------------
void VTKAppList::Initialize(void)
{
	VTKAppList::listing.clear();
	accessingMessageQueue = false;
	executeAppQueue = false;
	pthread_mutex_init(&messageMutex,NULL);
}

//----------------------
VTKApp** VTKAppList::AddApp(VTKAppProperties* props)
{
	WaitForExecuteAppAccess();
	//VTKApp* app = new VTKApp( x, y, width, height, name, appT, window);
	executeQueue.push_back(props); 
	props->ppApp = new VTKApp*;
	cerr << "App Pointer at: " << props->ppApp;
	*(props->ppApp) = NULL;
	VTKAppRelayStation* appRS = new VTKAppRelayStation();
	appRS->ppApp = props->ppApp;
	cerr << " -- Copied Over: " << appRS->ppApp << endl;
//	WaitForMessageQueueAccess();
	appRS->msgQueueHead = NULL;
	appRS->msgQueueTail = NULL;
//	FreeMessageQueueAccess();
	listing.push_back( appRS );
//	FreeExecuteAppAccess();
	executeAppQueue = true;
	cerr << "executeappqueue is true" << endl;
	NoAppRequested = false;
	return props->ppApp;
	
}

//------------------------
void BeginNewApps(void)
{


}

//----------------------
/*
void VTKAppList::KillApp(VTKApp *aNode)
{
	list<VTKApp** >::iterator itr = listing.begin();
	unsigned int ele = 0;
	while( ele < listing.size())
	{
		VTKApp** temp = *(itr);
		if( *(temp) == aNode)
		{
			listing.remove( temp );
			if( *(temp) != NULL)
			{
				VTKAppMessage* pMsg = new VTKAppMessage();
				pMsg->msg = VTKApp::VTKAPP_KILL;
				pMsg->msgData = NULL;
				(*(temp))->Message(pMsg);
				pMsg = NULL;
				
				while( (*(temp))->HaveMessages() )
				{
					//wait for the messages to be processed
				}
				
				delete (*temp);
				(*temp) = NULL;
				delete temp;
				
			}		
			
			return;
		}
		else
		{
			ele++;
			itr++;
		}
	}

	return;

}
*/
//----------------------
void VTKAppList::KillAllApps(void)
{
	VTKApp** temp = NULL;
	int ele = 0;
	list<VTKAppRelayStation* >::iterator itr = listing.begin();
	
	while( ele < listing.size() )
	{
		cerr << "Killing an app" << endl;
		temp = (*itr)->ppApp;
		//check to see if the exiting thread has already taken care of killing the app...just in case
		if((*temp) != NULL)
		{
			VTKAppMessage* pMsg = new VTKAppMessage();
			pMsg->msg = VTKApp::VTKAPP_KILL;
			pMsg->msgData = NULL;
			(*(temp))->Message(pMsg);
			pMsg = NULL;
		}
		
	
		if(*(temp))
		{
				
			delete (*(temp));
			(*temp) = NULL;
		}
		
	//	pthread_mutex_destroy(&(temp->mutexsum));
		delete temp;

		ele++;
		itr++;
	}
	
	listing.resize(0);
}

//------------------------
VTKApp* VTKAppList::GetApp(string aName )
{
	list<VTKAppRelayStation* >::iterator itr = listing.begin();
	unsigned int ele = 0;
	while( ele < listing.size())
	{
		VTKApp** temp = (*(itr))->ppApp;
		if( !strcmp( (*(temp))->GetAppName(), aName.c_str()) )
		{
			return (*(temp));
		}
		else
		{
			ele++;
			itr++;
		}
	}

	return NULL;
}

VTKApp* VTKAppList::GetApp(int index)
{
	list<VTKAppRelayStation* >::iterator itr = listing.begin();
	unsigned int ele = 0;
	while( ele < listing.size())
	{
		VTKApp** temp = (*(itr))->ppApp;
		if( ele == index )
		{
			return (*(temp));
		}
		else
		{
			ele++;
			itr++;
		}
	}

	return NULL;
}

int VTKAppList::GetNumApps(void)
{
	return listing.size();
}

void VTKAppList::Release(bool r)
{
	release = r;
}

bool VTKAppList::Release(void)
{
	return release;
}

bool VTKAppList::Freed(void)
{
	return freed;
}


void VTKAppList::RelayMessage(VTKAppMessage* msg, VTKApp** ppApp)
{
	
	list<VTKAppRelayStation* >::iterator itr = listing.begin();
	unsigned int ele = 0;
	while( ele < listing.size())
	{
		VTKApp** temp = (*(itr))->ppApp;
		if( temp == ppApp )
		{
			cerr << "Waiting for Access" << endl;
		//	WaitForMessageQueueAccess();
				VTKAppMessageQueue* queueNode = new VTKAppMessageQueue();
				queueNode->next = NULL;
				queueNode->pMsgStruct = msg;
				cerr << "Adding Message: " << msg << endl;
				msg = NULL;
				
				if((*itr)->msgQueueHead == NULL)
				{
				//	pthread_mutex_lock(&messageMutex);
					(*itr)->msgQueueHead = queueNode;
				//	pthread_mutex_unlock(&messageMutex);
					//(*itr)->msgQueueTail = queueNode;
				}
				else
				{
				//	pthread_mutex_lock( &messageMutex );
					(*(itr))->msgQueueTail->next = queueNode;
				//	pthread_mutex_unlock( &messageMutex );
				}
				
				
			//	pthread_mutex_lock( &messageMutex );
				(*(itr))->msgQueueTail = queueNode;
			//	pthread_mutex_unlock( &messageMutex );
		//	FreeMessageQueueAccess();
			return;
		}
		else
		{
			ele++;
			itr++;
		}
	}
	
	cerr << "You have been mislead! No app pointed by: " << ppApp << " exists" << endl;
	delete msg;
	return;
}

void VTKAppList::WaitForMessageQueueAccess()
{
//	cerr << "Waiting for Message Queue Access - ";
///	while( accessingMessageQueue );
//	accessingMessageQueue = true;
//	cerr << " Got Access!" << endl;
	pthread_mutex_lock( &messageMutex );
}

void VTKAppList::FreeMessageQueueAccess()
{
//	cerr << "Freeing Message Queue Access" << endl;
//	accessingMessageQueue = false;
	pthread_mutex_unlock( &messageMutex );
}

void VTKAppList::WaitForExecuteAppAccess()
{
//	cerr << "Waiting for App Execution Queue Access - ";
	while( executeAppQueue );
	executeAppQueue = true;
//	cerr << " Got Access!" << endl;
}

void VTKAppList::FreeExecuteAppAccess()
{
//	cerr << "Freeing App Execution Queue Acess" << endl;
	executeAppQueue = false;
}

void VTKAppList::TogglePauseAnimations()
{
	if(PauseAnims)
		PauseAnims = false;
	else
		PauseAnims = true;
}

/*
void* ThreadApp(void* data)
{
	VTKAppProperties* props = (VTKAppProperties* )data;
	VTKApp* app = new VTKApp( props->startx,
			 	  props->starty,
				  props->width,
				  props->height,
				  props->name,
				  props->appT,
				  NULL);
	app->viewThreadID = pthread_self();
	*(props->ppApp) = app;	  
	while(1)
	{
	
		if(app->HaveMessages())
			app->ProcessMessages();
	}
 
}*/
